import colorama
import difflib


def colorize(text, color):
    color_code = getattr(colorama.Fore, color.upper())
    return '%s%s%s' % (color_code, text, colorama.Fore.RESET)


def red(text):
    return colorize(text, 'red')


def green(text):
    return colorize(text, 'green')


def rangify(number_list):
    """Assumes the list is sorted."""
    if not number_list:
        return number_list

    ranges = []

    range_start = prev_num = number_list[0]
    for num in number_list[1:]:
        if num != (prev_num + 1):
            ranges.append((range_start, prev_num))
            range_start = num
        prev_num = num

    ranges.append((range_start, prev_num))
    return ranges


def extrapolate_coverage(lines_w_status):
    """
    Given the following input:

    >>> lines_w_status = [
        (1, True),
        (4, True),
        (7, False),
        (9, False),
    ]

    Return expanded lines with their extrapolated line status.

    >>> extrapolate_coverage(lines_w_status) == [
        (1, True),
        (2, True),
        (3, True),
        (4, True),
        (5, None),
        (6, None),
        (7, False),
        (8, False),
        (9, False),
    ]

    """
    lines = []

    prev_lineno = 0
    prev_status = True
    for lineno, status in lines_w_status:
        while (lineno - prev_lineno) > 1:
            prev_lineno += 1
            if prev_status is status:
                lines.append((prev_lineno, status))
            else:
                lines.append((prev_lineno, None))
        lines.append((lineno, status))
        prev_lineno = lineno
        prev_status = status

    return lines


def reconcile_lines(lines1, lines2):
    """
    Return a dict `{lineno1: lineno2}` which reconciles line numbers `lineno1`
    of list `lines1` to line numbers `lineno2` of list `lines2`. Only lines
    that are common in both sets are present in the dict, lines unique to one
    of the sets are omitted.
    """
    differ = difflib.Differ()
    diff = differ.compare(lines1, lines2)

    SAME = '  '
    ADDED = '+ '
    REMOVED = '- '
    INFO = '? '

    lineno_map = {}  # {lineno1: lineno2, ...}
    lineno1_offset = 0
    lineno2 = 1

    for diffline in diff:
        if diffline.startswith(INFO):
            continue

        if diffline.startswith(SAME):
            lineno1 = lineno2 + lineno1_offset
            lineno_map[lineno1] = lineno2

        elif diffline.startswith(ADDED):
            lineno1_offset -= 1

        elif diffline.startswith(REMOVED):
            lineno1_offset += 1
            continue

        lineno2 += 1

    return lineno_map



import lxml.etree as ET
import os

from pycobertura.utils import extrapolate_coverage, reconcile_lines


class Cobertura(object):
    """
    An XML cobertura parser.
    """
    def __init__(self, xml_source, base_path=None):
        """
        Initialize a Cobertura report given an XML file `xml_source` that is
        in the cobertura format.

        The optional argument `base_path` can be provided to resolve the path
        to the source code. `base_path` will prefix the filename attribute
        found in the Cobertura report.

        `xml_source` may be:
        - a path to an XML file
        - an open file object
        - an XML string
        """
        self.base_path = base_path
        self.source = xml_source
        self.xml = self._parse(xml_source)

    def _parse(self, xml_source):
        if hasattr(xml_source, 'startswith'):
            if xml_source.startswith('<?xml'):
                return self._parse_xml_string(xml_source)

            return self._parse_xml_file(xml_source)

        elif hasattr(xml_source, 'read'):
            return self._parse_xml_fileobj(xml_source)

    def _parse_xml_file(self, xml_file):
        return ET.parse(xml_file).getroot()

    def _parse_xml_fileobj(self, xml_fileobj):
        return self._parse_xml_string(xml_fileobj.read())

    def _parse_xml_string(self, xml_string):
        return ET.fromstring(xml_string)

    def _get_element_by_class_name(self, class_name):
        return self.xml.xpath("//class[@name='%s'][1]" % class_name)[0]

    def _get_lines_by_class_name(self, class_name):
        el = self._get_element_by_class_name(class_name)
        return el.xpath('./lines/line')

    @property
    def version(self):
        """Return the version number of the coverage report."""
        return self.xml.attrib['version']

    def line_rate(self, class_name=None):
        """
        Return the global line rate of the coverage report. If the class
        `class_name` is given, return the line rate of the class.
        """
        if class_name is None:
            el = self.xml
        else:
            el = self._get_element_by_class_name(class_name)

        return float(el.attrib['line-rate'])

    def branch_rate(self, class_name=None):
        """
        Return the global branch rate of the coverage report. If the class
        `class_name` is given, return the branch rate of the class.
        """
        if class_name is None:
            el = self.xml
        else:
            el = self._get_element_by_class_name(class_name)

        return float(el.attrib['branch-rate'])

    def missed_statements(self, class_name):
        """
        Return a list of uncovered line numbers for each of the missed
        statements found for the class `class_name`.
        """
        el = self._get_element_by_class_name(class_name)
        lines = el.xpath('./lines/line[@hits=0]')
        return [int(l.attrib['number']) for l in lines]

    def hit_statements(self, class_name):
        """
        Return a list of covered line numbers for each of the hit statements
        found for the class `class_name`.
        """
        el = self._get_element_by_class_name(class_name)
        lines = el.xpath('./lines/line[@hits>0]')
        return [int(l.attrib['number']) for l in lines]

    def line_statuses(self, class_name):
        """
        Return a list of tuples `(lineno, status)` of all the lines found in
        the Cobertura report where `lineno` is the line number and `status` is
        coverage status of the line which can be either `True` (line hit) or
        `False` (line miss).
        """
        line_elements = self._get_lines_by_class_name(class_name)

        lines_w_status = []
        for line in line_elements:
            lineno = int(line.attrib['number'])
            status = line.attrib['hits'] != '0'
            lines_w_status.append((lineno, status))

        return lines_w_status

    def missed_lines(self, class_name):
        """
        Return a list of extrapolated uncovered line numbers according to
        `Cobertura.line_statuses`.
        """
        statuses = self.line_statuses(class_name)
        statuses = extrapolate_coverage(statuses)
        return [lno for lno, status in statuses if status is False]

    def class_source(self, class_name):
        """
        Return a list of 3-element tuples `(lineno, source, status)` for each
        lines of code found in the source file of the class `class_name`.

        The 3 elements in each tuple are:
        `lineno`: line number in the source code
        `source`: actual source code for line number `lineno`
        `status`: True (hit) or False (miss)
        """
        filename = self.filename(class_name)

        if not os.path.exists(filename):
            return [(0, '%s not found' % filename, None)]

        with open(filename) as f:
            lines = []
            line_statuses = dict(self.line_statuses(class_name))
            for lineno, source in enumerate(f, start=1):
                line_status = line_statuses.get(lineno)
                lines.append((lineno, source, line_status))

        return lines

    def total_misses(self, class_name=None):
        """
        Return the total number of uncovered statements for the class
        `class_name`. If `class_name` is not given, return the total number of
        uncovered statements for all classes.
        """
        if class_name is not None:
            return len(self.missed_statements(class_name))

        total = 0
        for class_name in self.classes():
            total += len(self.missed_statements(class_name))

        return total

    def total_hits(self, class_name=None):
        """
        Return the total number of covered statements for the class
        `class_name`. If `class_name` is not given, return the total number of
        covered statements for all classes.
        """
        if class_name is not None:
            return len(self.hit_statements(class_name))

        total = 0
        for class_name in self.classes():
            total += len(self.hit_statements(class_name))

        return total

    def total_statements(self, class_name=None):
        """
        Return the total number of statements for the class `class_name`. If
        `class_name` is not given, return the total number of statements for
        all classes.
        """
        if class_name is not None:
            statements = self._get_lines_by_class_name(class_name)
            return len(statements)

        total = 0
        for class_name in self.classes():
            statements = self._get_lines_by_class_name(class_name)
            total += len(statements)

        return total

    def filename(self, class_name):
        """
        Return the filename of the class `class_name`. If `base_path` was
        provided in the constructor, it will be prefixed to the filename using
        `os.path.join`.
        """
        el = self._get_element_by_class_name(class_name)
        filename = el.attrib['filename']

        if self.base_path is None:
            return filename

        path = os.path.join(self.base_path, filename)
        return path

    def classes(self):
        """
        Return the list of available classes in the coverage report.
        """
        return [el.attrib['name'] for el in self.xml.xpath("//class")]

    def has_class(self, class_name):
        """
        Return `True` if the class `class_name` is present in the report,
        return `False` otherwise.
        """
        # FIXME: this will lookup a list which is slow, make it O(1)
        return class_name in self.classes()

    def packages(self):
        """
        Return the list of available packages in the coverage report.
        """
        return [el.attrib['name'] for el in self.xml.xpath("//package")]


class CoberturaDiff(object):
    """
    Diff Cobertura objects.
    """
    def __init__(self, cobertura1, cobertura2):
        self.cobertura1 = cobertura1
        self.cobertura2 = cobertura2

    def diff_total_statements(self, class_name=None):
        statements1 = self.cobertura1.total_statements(class_name)
        statements2 = self.cobertura2.total_statements(class_name)
        return statements2 - statements1

    def diff_total_misses(self, class_name=None):
        misses1 = self.cobertura1.total_misses(class_name)
        misses2 = self.cobertura2.total_misses(class_name)
        return misses2 - misses1

    def diff_total_hits(self, class_name=None):
        hits1 = self.cobertura1.total_hits(class_name)
        hits2 = self.cobertura2.total_hits(class_name)
        return hits2 - hits1

    def diff_line_rate(self, class_name=None):
        rate1 = self.cobertura1.line_rate(class_name)
        rate2 = self.cobertura2.line_rate(class_name)
        return rate2 - rate1

    def class_source(self, class_name):
        filename1 = self.cobertura1.filename(class_name)
        filename2 = self.cobertura2.filename(class_name)

        for filename in (filename1, filename2):
            if not os.path.exists(filename):
                return [(0, '%s not found' % filename, None)]

        try:
            f1 = open(filename1)
            f2 = open(filename2)
            lines1 = f1.readlines()
            lines2 = f2.readlines()
        finally:
            f1.close()
            f2.close()

        # Build a dict of lineno2 -> lineno1
        lineno_map = reconcile_lines(lines2, lines1)

        line_statuses1 = dict(self.cobertura1.line_statuses(class_name))
        line_statuses2 = dict(self.cobertura2.line_statuses(class_name))

        lines = []
        for lineno, line in enumerate(lines2, start=1):

            if lineno not in lineno_map:
                # line was added or removed, just use whatever coverage status
                # is available as there is nothing to compare against.
                line_status = line_statuses2.get(lineno)
            else:
                other_lineno = lineno_map[lineno]
                line_status1 = line_statuses1.get(other_lineno)
                line_status2 = line_statuses2.get(lineno)
                if line_status1 is line_status2:
                    line_status = None  # unchanged
                elif line_status1 is True and line_status2 is False:
                    line_status = False  # decreased
                elif line_status1 is False and line_status2 is True:
                    line_status = True  # increased

            lines.append((lineno, line, line_status))

        return lines



from .utils import make_cobertura


def test_diff_class_source():
    from pycobertura.cobertura import CoberturaDiff

    cobertura1 = make_cobertura(
        'tests/dummy.source1.xml',
        base_path='tests/dummy.source1/'
    )
    cobertura2 = make_cobertura(
        'tests/dummy.source2.xml',
        base_path='tests/dummy.source2/'
    )
    differ = CoberturaDiff(cobertura1, cobertura2)

    expected_sources = {
        'dummy/__init__': [],
        'dummy/dummy': [
            (1, 'def foo():\n', None),
            (2, '    pass\n', None),
            (3, '\n', None),
            (4, 'def bar():\n', None),
            (5, "    a = 'a'\n", True),
            (6, "    d = 'd'\n", True)
        ],
        'dummy/dummy2': [
            (1, 'def baz():\n', None),
            (2, "    c = 'c'\n", True),
            (3, '\n', None),
            (4, 'def bat():\n', True),
            (5, '    pass\n', False)
        ],
    }

    for class_name in cobertura2.classes():
        assert differ.class_source(class_name) == \
            expected_sources[class_name]


def test_diff_total_misses():
    from pycobertura.cobertura import CoberturaDiff

    cobertura1 = make_cobertura(
        'tests/dummy.source1.xml',
        base_path='tests/dummy.source1/'
    )
    cobertura2 = make_cobertura(
        'tests/dummy.source2.xml',
        base_path='tests/dummy.source2/'
    )
    differ = CoberturaDiff(cobertura1, cobertura2)

    expected_sources = {
        'dummy/__init__': 0,
        'dummy/dummy': -2,
        'dummy/dummy2': 1,
    }

    assert differ.diff_total_misses() == -1


def test_diff_total_misses_by_class():
    from pycobertura.cobertura import CoberturaDiff

    cobertura1 = make_cobertura(
        'tests/dummy.source1.xml',
        base_path='tests/dummy.source1/'
    )
    cobertura2 = make_cobertura(
        'tests/dummy.source2.xml',
        base_path='tests/dummy.source2/'
    )
    differ = CoberturaDiff(cobertura1, cobertura2)

    expected_sources = {
        'dummy/__init__': 0,
        'dummy/dummy': -2,
        'dummy/dummy2': 1,
    }

    for class_name in cobertura2.classes():
        assert differ.diff_total_misses(class_name) == \
            expected_sources[class_name]


def test_diff_line_rate():
    from pycobertura.cobertura import CoberturaDiff

    cobertura1 = make_cobertura(
        'tests/dummy.source1.xml',
        base_path='tests/dummy.source1/'
    )
    cobertura2 = make_cobertura(
        'tests/dummy.source2.xml',
        base_path='tests/dummy.source2/'
    )
    differ = CoberturaDiff(cobertura1, cobertura2)

    assert differ.diff_line_rate() == 0.17459999999999998


def test_diff_line_rate_by_class():
    from pycobertura.cobertura import CoberturaDiff

    cobertura1 = make_cobertura(
        'tests/dummy.source1.xml',
        base_path='tests/dummy.source1/'
    )
    cobertura2 = make_cobertura(
        'tests/dummy.source2.xml',
        base_path='tests/dummy.source2/'
    )
    differ = CoberturaDiff(cobertura1, cobertura2)

    expected_sources = {
        'dummy/__init__': 0,
        'dummy/dummy': 0.4,
        'dummy/dummy2': -0.25,
    }

    for class_name in cobertura2.classes():
        assert differ.diff_line_rate(class_name) == \
            expected_sources[class_name]


def test_diff_total_hits():
    from pycobertura.cobertura import CoberturaDiff

    cobertura1 = make_cobertura(
        'tests/dummy.source1.xml',
        base_path='tests/dummy.source1/'
    )
    cobertura2 = make_cobertura(
        'tests/dummy.source2.xml',
        base_path='tests/dummy.source2/'
    )
    differ = CoberturaDiff(cobertura1, cobertura2)

    assert differ.diff_total_hits() == 3


def test_diff_total_hits_by_class():
    from pycobertura.cobertura import CoberturaDiff

    cobertura1 = make_cobertura(
        'tests/dummy.source1.xml',
        base_path='tests/dummy.source1/'
    )
    cobertura2 = make_cobertura(
        'tests/dummy.source2.xml',
        base_path='tests/dummy.source2/'
    )
    differ = CoberturaDiff(cobertura1, cobertura2)

    expected_total_hits = {
        'dummy/__init__': 0,
        'dummy/dummy': 2,
        'dummy/dummy2': 1,
    }

    for class_name in cobertura2.classes():
        assert differ.diff_total_hits(class_name) == \
            expected_total_hits[class_name]



import click

from pycobertura.cobertura import Cobertura
from pycobertura.reporters import (
    HtmlReporter, TextReporter, HtmlReporterDelta, TextReporterDelta
)


pycobertura = click.Group()


reporters = {
    'html': HtmlReporter,
    'text': TextReporter,
}


@pycobertura.command()
@click.argument('cobertura_file')
@click.option(
    '-f', '--format', default='text',
    type=click.Choice(list(reporters))
)
@click.option(
    '-o', '--output', metavar='<file>', type=click.File('wb'),
    help='Write output to <file> instead of stdout.'
)
@click.option(
    '-s', '--source', metavar='<source-dir>',
    help='Provide path to source code directory for HTML output.'
)
def show(cobertura_file, format, output, source):
    """show coverage summary of a Cobertura report"""
    cobertura = Cobertura(cobertura_file, base_path=source)
    Reporter = reporters[format]
    reporter = Reporter(cobertura)
    report = reporter.generate()

    if not isinstance(report, bytes):
        report = report.encode('utf-8')

    isatty = True if output is None else output.isatty()
    click.echo(report, file=output, nl=isatty)


delta_reporters = {
    'text': TextReporterDelta,
    'html': HtmlReporterDelta,
}


@pycobertura.command(help="""\
The diff command compares and shows the changes between two Cobertura reports.

Missing lines with the diff command can only be accurately computed if the
source code to each of the coverage reports is given using --source1 and
--source2. Not providing the path to the source code assumes that each of the
coverage files were generated against the same source code which is accessible
by looking at the "filename" attribute in the Cobertura reports as if --source1
and --source2 options were given the same path to the source.
""")
@click.argument('cobertura_file1')
@click.argument('cobertura_file2')
@click.option(
    '--color/--no-color', default=None,
    help='Colorize the output. By default, pycobertura emits color codes only '
         'when standard output is connected to a terminal. This has no effect '
         'with the HTML output format.')
@click.option(
    '-f', '--format', default='text',
    type=click.Choice(list(delta_reporters))
)
@click.option(
    '-o', '--output', metavar='<file>', type=click.File('wb'),
    help='Write output to <file> instead of stdout.'
)
@click.option(
    '-s1', '--source1', metavar='<source-dir1>',
    help='Provide path to first source code directory.'
)
@click.option(
    '-s2', '--source2', metavar='<source-dir2>',
    help='Provide path to second source code directory.'
)
def diff(
        cobertura_file1, cobertura_file2,
        color, format, output, source1, source2):
    """compare coverage of two Cobertura reports"""
    cobertura1 = Cobertura(cobertura_file1, base_path=source1)
    cobertura2 = Cobertura(cobertura_file2, base_path=source2)

    Reporter = delta_reporters[format]
    reporter_args = [cobertura1, cobertura2]
    reporter_kwargs = {}

    isatty = True if output is None else output.isatty()

    if format == 'text':
        color = isatty if color is None else color is True
        reporter_kwargs['color'] = color

    reporter = Reporter(*reporter_args, **reporter_kwargs)
    report = reporter.generate()

    if not isinstance(report, bytes):
        report = report.encode('utf-8')

    # FIXME: In Click 3.x, the color is stripped out if the output file is
    # detected to not support ANSI codes causing `-o outfile --color` to write
    # to `outfile` without colors. In Click 4.x, `echo()` will take a
    # color=True/False flag which will allow us to override ANSI code
    # auto-detection.
    # https://github.com/mitsuhiko/click/commit/5cf7f2ddfba328070751cbda32782520d4e0d6f5
    # click.echo(report, file=output, nl=isatty, color=color)  # click 4.x
    click.echo(report, file=output, nl=isatty)

    # non-zero exit code if line rate worsened
    exit_code = cobertura1.line_rate() > cobertura2.line_rate()
    raise SystemExit(exit_code)
